#! /usr/bin/env python
###############################################################################
#
# Copyright (C) 2014
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

"""Test case for unb1_test

Usage:

   --rep = number of intervals that diagnostics results are verified
   --sim targets a running simulation.

Description:
   This test case tests:
   - system info
   - read sensors
   - read ppsh
   - write to wdi to force reload from bank 0
   - flash access: write image to bank 1
   - remote update: start image in bank 1
   - read status MAC 10GbE
   - read status PHY XAUI
   - BG - DB tests for 1 port on 1 FPGA
   - BG - DB tests for all ports on all FPGAs
   - link test between ports on the same FPGA (same XO)
   - link test between ports on different Uniboards (XO drift)
   - stable link after image restart

"""

###############################################################################
# System imports
import sys
import test_case
import node_io
import pi_system_info
import pi_unb_sens
import pi_ppsh
import pi_wdi
import pi_epcs
import pi_remu
import pi_eth
import pi_bsn_monitor
import pi_diag_block_gen
import pi_diag_data_buffer

from tools import *
from common import *
from pi_common import *


def show_compare_arrays(tc,a,b):
    import numpy as np
    tc.append_log(6, 'Compare array A:\n%s' % np.asarray(a))
    tc.append_log(6, 'With array B:\n%s' % np.asarray(b))
    compared = (np.asarray(a) == np.asarray(b))
    tc.append_log(5, 'Result:\n%s' % compared)
    return (not False in compared)

def test_BG_to_DB(tc,io,cmd,use_pps=False):
    tc.set_section_id('Test BG - ')
    tc.append_log(3, '>>>')
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    tc.append_log(3, '>>>')

    if use_pps==True: ppsh = pi_ppsh.PiPpsh(tc, io)

    tc.set_section_id('Read - ')
    nof_streams=3
    blocksize=0
    Bg = pi_diag_block_gen.PiDiagBlockGen(tc,io,nofChannels=nof_streams,ramSizePerChannel=blocksize)
    Bg.write_disable()

    settings = Bg.read_block_gen_settings()
    samples_per_packet = settings[0][1]
    gapsize            = settings[0][3]
    blocksize          = pow(2, ceil_log2(samples_per_packet+gapsize))

    Bg = pi_diag_block_gen.PiDiagBlockGen(tc,io,nofChannels=nof_streams, ramSizePerChannel=blocksize)
    #Bg.write_block_gen_settings(samplesPerPacket=700, blocksPerSync=781250, gapSize=300, memLowAddr=0, memHighAddr=701, BSNInit=42)
    Db = pi_diag_data_buffer.PiDiagDataBuffer(tc,io,nofStreams=nof_streams,ramSizePerStream=blocksize)
    resetptrn = [0xc1ea1ed1]*blocksize #samples_per_packet + [0]*(blocksize-samples_per_packet)
    for s in tc.spNrs:
        Db.overwrite_data_buffer(resetptrn,streamNr=s,vLevel=9)

    db_ram = []
    for s in tc.spNrs:
        db_ram.append(Db.read_data_buffer(streamNr=s,vLevel=5))

    if use_pps==True:
        Bg.write_enable_pps()
        ppsh.wait_until_pps(vLevel=6)
    else:
        Bg.write_enable()


    bg_ram = []
    for s in tc.spNrs:
        ram = Bg.read_waveform_ram(channelNr=s,vLevel=5)
        rram=[]
        for r in ram: rram.append(list(r)) # () -> []
        bg_ram.append(rram)


    db_ram = []
    for s in tc.spNrs:
        db_ram.append(Db.read_data_buffer(streamNr=s,vLevel=5))


    # truncate: in 10GbE the data is only sized 700
    for s in tc.spNrs:
        bg_ram[s][0] = bg_ram[s][0][:-(blocksize-700)]
        db_ram[s][0] = db_ram[s][0][:-(blocksize-700)]
        #bg_ram[s][0] = bg_ram[s][0][:-(blocksize-samples_per_packet)]
        #db_ram[s][0] = db_ram[s][0][:-(blocksize-samples_per_packet)]

    #print 'bg_ram=',bg_ram
    #print 'db_ram=',db_ram

    tc.set_section_id('Compare (BG==DB) - ')
    compared = (bg_ram == db_ram)
    tc.append_log(3, '')
    tc.append_log(1, '>>> (BG==DB): %s' % compared)
    tc.append_log(3, '')
    if compared==False: tc.set_result('FAILED')
    show_compare_arrays(tc,bg_ram,db_ram)
    Bg.write_disable()

def test_BG_to_DB_pps(tc,io,cmd):
    return test_BG_to_DB(tc,io,cmd,use_pps=True)



def test_tr_xaui(tc,io,cmd):
    tc.set_section_id('Read tr xaui status - ')
    tc.append_log(3, '>>>')
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    tc.append_log(3, '>>>')

    tc.append_log(3, '')
    from pi_tr_xaui import REGMAP,PiTrXaui
    on_execute(class_definition=PiTrXaui,regmap=REGMAP)



def test_tr_10GbE(tc,io,cmd):
    tc.set_section_id('Read tr 10GbE status - ')
    tc.append_log(3, '>>>')
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    tc.append_log(3, '>>>')

    tc.append_log(3, '')
    from pi_tr_10GbE import REGMAP,PiTr10GbE
    on_execute(class_definition=PiTr10GbE,regmap=REGMAP)



def test_bsn_mon(tc,io,cmd):
    tc.set_section_id('Read BSN monitor status - ')
    tc.append_log(3, '>>>')
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    tc.append_log(3, '>>>')

    Bsn = pi_bsn_monitor.PiBsnMonitor(tc,io)
    Bsn.read_bsn_monitor()
    tc.append_log(3, '')



def test_info(tc,io,cmd):
    tc.set_section_id('Read System Info - ')
    tc.append_log(3, '>>>')
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    tc.append_log(3, '>>>')

    #with node_io.NodeIO(tc.nodeImages, tc.base_ip) as temp_io:
    info = pi_system_info.PiSystemInfo(tc, io)
    tc.append_log(1, '>>> reading REGMAPs')
    info.make_register_info()
    tc.append_log(3, '')
    info.read_system_info()
    tc.append_log(3, '')
    info.read_use_phy()
    tc.append_log(3, '')
    design_name = info.read_design_name()
    tc.append_log(1, '>>> design_name=%s' % design_name)
    tc.append_log(3, '')
    info.read_stamps()
    tc.append_log(3, '')
    info.read_design_note()
    
    expected_design_name = tc.gpString
    if cmd==3 and expected_design_name != '':
            tc.set_section_id('Verify System Info - ')
            tc.append_log(1, '>>> Verifying design_name == %s' % expected_design_name)
            for name in design_name:
                if (name != expected_design_name): 
                    tc.set_result('FAILED')
                    tc.append_log(2, '>>> design_name mismatch!! (%s != %s)' % (name,expected_design_name))


    
def test_sensors(tc,io,cmd):
    tc.set_section_id('Read sensors - ')
    tc.append_log(3, '>>>')
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    tc.append_log(3, '>>>')
    
    sens = pi_unb_sens.PiUnbSens(tc, io)

    sens.read_unb_sensors()
    tc.append_log(3, '')
    sens.read_fpga_temperature()
    tc.append_log(3, '')
    sens.read_eth_temperature()
    tc.append_log(3, '')
    sens.read_unb_current()
    sens.read_unb_voltage()
    sens.read_unb_power()



def test_ppsh(tc,io,cmd):
    tc.set_section_id('Read PPSH capture count - ')
    tc.append_log(3, '>>>')
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    tc.append_log(3, '>>>')
    
    Ppsh = pi_ppsh.PiPpsh(tc, io)
    Ppsh.read_ppsh_capture_cnt()
    tc.append_log(3, '')



def test_wdi(tc,io,cmd):
    tc.set_section_id('Reset to image in bank 0 using WDI - ')
    tc.append_log(3, '>>>')
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    tc.append_log(3, '>>>')

    Wdi = pi_wdi.PiWdi(tc, io)
    Wdi.write_wdi_override()
    tc.append_log(3, '')



def test_remu(tc,io,cmd):
    tc.set_section_id('REMU start image in bank 1 - ')
    tc.append_log(3, '>>>')
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    tc.append_log(3, '>>>')

    dummy_tc = test_case.Testcase('Dummy TB - ', '')
    dummy_tc.set_result('PASSED')
    
    #with pi_remu.PiRemu(dummy_tc, io) as Remu:
    Remu = pi_remu.PiRemu(dummy_tc, io)
    try:
            Remu.write_user_reconfigure()
    except:
            pass # ignoring FAILED

    if dummy_tc.get_result() == 'FAILED':
            tc.append_log(1, 'Result=%s but ignoring this' % dummy_tc.get_result())

    tc.append_log(3, '')



def test_eth(tc,io,cmd):
    tc.set_section_id('ETH status - ')
    tc.append_log(3, '>>>')
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    tc.append_log(3, '>>>')
    
    eth = pi_eth.PiEth(tc, io)
    hdr=eth.read_hdr(0)
    eth.disassemble_hdr(hdr)
    tc.append_log(3, '')



def test_flash(tc,io,cmd):
    tc.set_section_id('Flash write to bank 1 - ')
    tc.append_log(3, '>>>')
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    tc.append_log(3, '>>>')
    
    Epcs = pi_epcs.PiEpcs(tc, io)
    path_to_rbf = instanceName = tc.gpString
    Epcs.write_raw_binary_file("user", path_to_rbf)
    tc.append_log(3, '')

    tc.set_section_id('Flash read/verify bank 1 - ')
    tc.append_log(3, '>>>')
    tc.append_log(1, '>>> Read from flash (pi_epcs.py)')
    tc.append_log(3, '>>>')
    
    path_to_rbf = instanceName = tc.gpString
    Epcs.read_and_verify_raw_binary_file("user", path_to_rbf)
    tc.append_log(3, '')



def sleep(tc,io,cmd):
    tc.set_section_id('%s - ' % cmd_text(cmd))
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    if cmd == 0:
        tc.sleep(1.0)
    elif cmd == 1:
        tc.sleep(5.0)

def show_help(tc,io,cmd):
    help_text(tc,io,100)

def show_example(tc,io,cmd):
    tc.append_log(1, '>>> %s' % help_text(tc,io,cmd))
    help_text(tc,io,101)


# Avaliable commands
Cmd = {
    0 : ('SLEEP1' , sleep,        'Sleep 1 second',''),
    1 : ('SLEEP5' , sleep,        'Sleep 5 seconds',''),
    2 : ('INFO'   , test_info,    'using pi_system_info to read system info (access PIO_SYSTEM_INFO)',''),
    3 : ('INFO+'  , test_info,    'using pi_system_info to read system info (access PIO_SYSTEM_INFO)','(-s for expected design_name)'),
    4 : ('FLASH'  , test_flash,   'using pi_epcs to program/verify flash','(-s for .rbf file)'),
    5 : ('SENSORS', test_sensors, 'using pi_unb_sens to readout sensors (access REG_UNB_SENS)',''),
    6 : ('PPSH'   , test_ppsh,    'using pi_ppsh to read PPSH capture count (access PIO_PPS)',''),
    7 : ('ETH'    , test_eth,     'using pi_eth to read eth status',''),
    8 : ('REMU'   , test_remu,    'using pi_remu to load user image (access REG_REMU)',''),
    9 : ('WDI'    , test_wdi,     'using pi_wdi to reset to image in bank 0 (access REG_WDI)',''),
   10 : ('XAUI'   , test_tr_xaui, 'using pi_tr_xaui to read xaui status (access REG_TR_XAUI)',''),
   11 : ('10GBE'  , test_tr_10GbE,'using pi_tr_10GbE to read 10GbE status (access REG_TR_10GBE)',''),
   12 : ('BSN'    , test_bsn_mon, 'using pi_bsn_monitor to read BSN monitor (access REG_BSN_MONITOR)',''),
   13 : ('BGDB'   , test_BG_to_DB,'using BG (pi_diag_block_gen.py) and DB (pi_diag_data_buffer.py)','(--sp for addressing streams)'),
   14 : ('BGDBpps', test_BG_to_DB_pps,'using BG (pi_diag_block_gen.py) and DB (pi_diag_data_buffer.py) using PPS','(--sp for addressing streams)'),
   15 : ('EXAMPLE', show_example, 'show several example commands',''),
  100 : ('HELP'   , show_help,    'show help on commands','')
}

def cmd_text(cmd):
    return Cmd[cmd][0]

def help_text(tc,io,cmd):
    str=''
    if cmd == 100:
        tc.append_log(0, '>>> Usage [--cmd <command number or series of numbers>]:')
        for cmd in sorted(Cmd):
            tc.append_log(0, '    .  %d : %s: %s : %s' % (cmd,Cmd[cmd][0],Cmd[cmd][2],Cmd[cmd][3]))
    elif cmd == 101: # examples
        tc.append_log(0, '')
        tc.append_log(0, '[reset, load user img] sequence: --cmd 2,9,1,2,9,1,2,8,1,2')
        #                                                          ^fact ^fact ^user
        tc.append_log(0, '[flash+start user img] sequence: --cmd 2,4,9,1,2,8,1,2 -s file.rbf')
        #                                                            ^fact ^user
        tc.append_log(0, '[re-read info,sensors] sequence: --cmd 3,5,6 --rep 10 -s expected_design_name')
        tc.append_log(0, '[reset to factory]     sequence: --cmd 2,9,1,2')
        tc.append_log(0, '[program user image]   sequence: --cmd 4 -s file.rbf')
        tc.append_log(0, '[load user image]      sequence: --cmd 8,1,2')
    else:
        str = Cmd[cmd][2]
    return str




##################################################################################################################
# Main
#
# Create a test case object
tc = test_case.Testcase('TB - ', '')
tc.set_result('PASSED')
dgnName = tc.gpString
tc.append_log(3, '>>>')
tc.append_log(1, '>>> Title : Test bench for tc_unb1_test.py on nodes %s, %s' % (tc.unb_nodes_string(''),dgnName))
tc.append_log(3, '>>>')
    

# Create access object for nodes
#io = node_io.NodeIO(tc.nodeImages, tc.base_ip)



##################################################################################################################
# Run tests
rep = 0
while rep < tc.repeat:
    rep += 1
    tc.append_log(3, '')
    tc.append_log(2, 'Rep = %d' % rep)
    
    for cmd in tc.commands:
        if cmd > len(Cmd): cmd = 100 # default to help_text
        tc.append_log(3, 'Next command: %d : %s: %s' % (cmd,Cmd[cmd][0],Cmd[cmd][2]))

        with node_io.NodeIO(tc.nodeImages, tc.base_ip) as temp_io:
            Cmd[cmd][1](tc,temp_io,cmd)




##################################################################################################################
# End
tc.set_section_id('')
tc.append_log(3, '')
tc.append_log(3, '>>>')
tc.append_log(0, '>>> Test bench result: %s' % tc.get_result())
tc.append_log(3, '>>>')

sys.exit(tc.get_result())

