-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, tech_tse_lib, technology_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_network_total_header_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE tech_tse_lib.tech_tse_pkg.ALL;
USE work.eth_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

-- Purpose:
--   Provide Ethernet control access to a node and some UDP ports for streaming
--   data.
-- Description:
--   Connect the 1GbE TSE to the microprocessor and to streaming UDP ports. The
--   packets for the streaming channels are directed based on the UDP port
--   number and all other packets are transfered to the default control channel.

ENTITY eth IS
  GENERIC (
    g_technology         : NATURAL := c_tech_select_default;
    g_cross_clock_domain : BOOLEAN := TRUE;    -- use FALSE when mm_clk and st_clk are the same, else use TRUE to cross the clock domain
    g_ETH_PHY            : STRING  := "LVDS";  -- "LVDS" (default): uses LVDS IOs for ctrl_unb_common, "XCVR": uses tranceiver PHY
    g_ihl20              : BOOLEAN := FALSE
  );
  PORT (
    -- Clocks and reset
    mm_rst             : IN  STD_LOGIC;  -- reset synchronous with mm_clk
    mm_clk             : IN  STD_LOGIC;  -- memory-mapped bus clock
    eth_clk            : IN  STD_LOGIC;  -- ethernet phy reference clock
    st_rst             : IN  STD_LOGIC;  -- reset synchronous with st_clk
    st_clk             : IN  STD_LOGIC;  -- packet stream clock
    
    cal_rec_clk        : IN  STD_LOGIC := '0';  -- Calibration & reconfig clock when using XCVR
    
    -- UDP transmit interface
    udp_tx_snk_in_arr  : IN  t_dp_sosi_arr(c_eth_nof_udp_ports-1 DOWNTO 0) := (OTHERS=> c_dp_sosi_rst);  -- ST sinks, default not valid if not used
    udp_tx_snk_out_arr : OUT t_dp_siso_arr(c_eth_nof_udp_ports-1 DOWNTO 0);
    -- UDP receive interface
    udp_rx_src_in_arr  : IN  t_dp_siso_arr(c_eth_nof_udp_ports-1 DOWNTO 0) := (OTHERS=> c_dp_siso_rdy);  -- ST sources, default ready if not used
    udp_rx_src_out_arr : OUT t_dp_sosi_arr(c_eth_nof_udp_ports-1 DOWNTO 0);

    -- Memory Mapped Slaves
    tse_sla_in         : IN  t_mem_mosi;  -- ETH TSE MAC registers
    tse_sla_out        : OUT t_mem_miso;
    reg_sla_in         : IN  t_mem_mosi;  -- ETH control and status registers
    reg_sla_out        : OUT t_mem_miso;
    reg_sla_interrupt  : OUT STD_LOGIC;   -- Interrupt
    ram_sla_in         : IN  t_mem_mosi;  -- ETH rx frame and tx frame memory
    ram_sla_out        : OUT t_mem_miso;

    -- PHY interface
    eth_txp            : OUT STD_LOGIC;
    eth_rxp            : IN  STD_LOGIC;

    -- LED interface
    tse_led            : OUT t_tech_tse_led
  );
END eth;


ARCHITECTURE str OF eth IS

  ------------------------------------------------------------------------------
  -- ETH Rx packet buffer and Tx packet buffer
  ------------------------------------------------------------------------------
  
  -- Use MM bus data width = c_word_w = 32
  CONSTANT c_mm_ram  : t_c_mem := (latency  => c_mem_ram_rd_latency,
                                   adr_w    => c_eth_ram_addr_w,
                                   dat_w    => c_word_w,
                                   nof_dat  => c_eth_ram_nof_words,
                                   init_sl  => '0');
                                   
  SIGNAL mem_in         : t_mem_mosi;  -- big endian on ST and TSE MAC network side
  SIGNAL mem_out        : t_mem_miso;  -- big endian on ST and TSE MAC network side
  SIGNAL mem_in_endian  : t_mem_mosi;  -- keep big endian on MM side
  SIGNAL mem_out_endian : t_mem_miso;  -- keep big endian on MM side
                                   
  ------------------------------------------------------------------------------
  -- ETH stream
  ------------------------------------------------------------------------------
  
  -- Multiplex - demultiplex
  CONSTANT c_mux_nof_ports    : NATURAL := 1 + c_eth_nof_udp_ports; -- One for control + nof UDP ports
  CONSTANT c_demux_nof_ports  : NATURAL := c_mux_nof_ports;
  CONSTANT c_demux_combined   : BOOLEAN := FALSE;  -- when TRUE then all downstream sinks must be ready, when FALSE then only the
                                                   -- selected sink needs to be ready (see dp_demux for more explanation).
  -- All Rx (so UDP off-load and other ETH traffic)
  SIGNAL rx_adapt_siso        : t_dp_siso;
  SIGNAL rx_adapt_sosi        : t_dp_sosi;
  
  SIGNAL rx_crc_siso          : t_dp_siso;
  SIGNAL rx_crc_sosi          : t_dp_sosi;
  
  SIGNAL rx_ihl20_siso        : t_dp_siso;
  SIGNAL rx_ihl20_sosi        : t_dp_sosi;
  
  SIGNAL rx_hdr_siso          : t_dp_siso;
  SIGNAL rx_hdr_sosi          : t_dp_sosi;
  SIGNAL rx_hdr_status        : t_eth_hdr_status;
  
  -- Rx demux
  SIGNAL rx_channel_siso      : t_dp_siso;
  SIGNAL rx_channel_sosi      : t_dp_sosi;
  
  SIGNAL demux_siso_arr       : t_dp_siso_arr(0 TO c_mux_nof_ports-1);
  SIGNAL demux_sosi_arr       : t_dp_sosi_arr(0 TO c_mux_nof_ports-1);
  
  -- ETH Rx
  SIGNAL eth_rx_siso          : t_dp_siso;
  SIGNAL eth_rx_sosi          : t_dp_sosi;
  
  SIGNAL rx_frame_rd            : STD_LOGIC;
  SIGNAL rx_frame_ack           : STD_LOGIC;
  SIGNAL rx_frame_done          : STD_LOGIC;
  SIGNAL rx_frame_sosi          : t_dp_sosi;
  SIGNAL rx_frame_hdr_words_arr : t_network_total_header_32b_arr;
  SIGNAL rx_frame_hdr_fields    : t_network_total_header;
  SIGNAL rx_frame_hdr_status    : t_eth_hdr_status;
  SIGNAL rx_frame_crc_word      : STD_LOGIC_VECTOR(c_eth_data_w-1 DOWNTO 0);
  
  -- ETH Tx
  SIGNAL eth_tx_siso          : t_dp_siso;
  SIGNAL eth_tx_sosi          : t_dp_sosi;
  
  -- Tx mux
  SIGNAL mux_siso_arr         : t_dp_siso_arr(0 TO c_mux_nof_ports-1);
  SIGNAL mux_sosi_arr         : t_dp_sosi_arr(0 TO c_mux_nof_ports-1);
  
  -- All Tx (so UDP off-load and other ETH traffic)
  SIGNAL tx_mux_siso          : t_dp_siso;
  SIGNAL tx_mux_sosi          : t_dp_sosi;
  
  SIGNAL tx_hdr_siso          : t_dp_siso;
  SIGNAL tx_hdr_sosi          : t_dp_sosi;
  
  ------------------------------------------------------------------------------
  -- MM registers (in st_clk domain)
  ------------------------------------------------------------------------------
  
  -- . write/read back
  SIGNAL reg_demux            : t_eth_mm_reg_demux;
  SIGNAL reg_config           : t_eth_mm_reg_config;
  SIGNAL reg_control          : t_eth_mm_reg_control;
  SIGNAL reg_continue_wr      : STD_LOGIC;
  -- . read only
  SIGNAL reg_frame            : t_eth_mm_reg_frame;
  SIGNAL reg_status           : t_eth_mm_reg_status;
  SIGNAL reg_status_wr        : STD_LOGIC;
  
  ------------------------------------------------------------------------------
  -- TSE MAC
  ------------------------------------------------------------------------------
  -- . MAC Transmit Stream
  SIGNAL tse_tx_siso        : t_dp_siso;
  SIGNAL tse_tx_sosi        : t_dp_sosi;
  -- . MAC specific
  SIGNAL tse_tx_mac_in      : t_tech_tse_tx_mac;
  SIGNAL tse_tx_mac_out     : t_tech_tse_tx_mac;
  -- . MAC Receive Stream
  SIGNAL tse_rx_sosi        : t_dp_sosi;
  SIGNAL tse_rx_siso        : t_dp_siso;
  -- . MAC specific
  SIGNAL tse_rx_mac_out     : t_tech_tse_rx_mac;

BEGIN

  ------------------------------------------------------------------------------
  -- MM registers
  ------------------------------------------------------------------------------
  
  u_mm_registers : ENTITY work.eth_mm_registers
  GENERIC MAP (
    g_cross_clock_domain => g_cross_clock_domain
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst             => mm_rst,
    mm_clk             => mm_clk,
    st_rst             => st_rst,
    st_clk             => st_clk,
    -- Memory Mapped Slave
    sla_in             => reg_sla_in,
    sla_out            => reg_sla_out,
    sla_interrupt      => reg_sla_interrupt,
    -- MM registers in st_clk domain
    -- . write/read back
    st_reg_demux       => reg_demux,
    st_reg_config      => reg_config,
    st_reg_control     => reg_control,
    st_reg_continue_wr => reg_continue_wr,
    -- . read only
    st_reg_frame       => reg_frame,
    st_reg_status      => reg_status,
    st_reg_status_wr   => reg_status_wr
  );

  -- Packet buffer
  u_mm_ram : ENTITY common_lib.common_ram_crw_crw
  GENERIC MAP (
    g_ram     => c_mm_ram
  )
  PORT MAP (
    rst_a     => mm_rst,
    clk_a     => mm_clk,
    wr_en_a   => ram_sla_in.wr,
    adr_a     => ram_sla_in.address(c_mm_ram.adr_w-1 DOWNTO 0),
    wr_dat_a  => ram_sla_in.wrdata(c_mm_ram.dat_w-1 DOWNTO 0),
    rd_en_a   => ram_sla_in.rd,
    rd_dat_a  => ram_sla_out.rddata(c_mm_ram.dat_w-1 DOWNTO 0),
    rd_val_a  => ram_sla_out.rdval,
    rst_b     => st_rst,
    clk_b     => st_clk,
    wr_en_b   => mem_in_endian.wr,
    adr_b     => mem_in_endian.address(c_mm_ram.adr_w-1 DOWNTO 0),
    wr_dat_b  => mem_in_endian.wrdata(c_mm_ram.dat_w-1 DOWNTO 0),
    rd_en_b   => mem_in_endian.rd,
    rd_dat_b  => mem_out_endian.rddata(c_mm_ram.dat_w-1 DOWNTO 0),
    rd_val_b  => mem_out_endian.rdval
  );
  
  -- The Rx, Tx packet buffer is big-endian
  mem_in_endian <= func_mem_swap_endianess(mem_in, c_word_sz);
  mem_out       <= func_mem_swap_endianess(mem_out_endian, c_word_sz);
  
  
  ------------------------------------------------------------------------------
  -- RX : Adapt the TSE RX source ready latency from 2 to 1
  ------------------------------------------------------------------------------
  
  u_adapt : ENTITY dp_lib.dp_latency_adapter
  GENERIC MAP (
    g_in_latency  => c_eth_rx_ready_latency,  -- = 2
    g_out_latency => c_eth_ready_latency      -- = 1
  )
  PORT MAP (
    rst     => st_rst,
    clk     => st_clk,
    -- ST sink
    snk_out => tse_rx_siso,
    snk_in  => tse_rx_sosi,
    -- ST source
    src_in  => rx_adapt_siso,
    src_out => rx_adapt_sosi
  );
     
  ------------------------------------------------------------------------------
  -- RX : Replace the CRC word with the stream error field from the TSE MAC
  ------------------------------------------------------------------------------
  
  u_crc_ctrl : ENTITY work.eth_crc_ctrl
  PORT MAP (
    rst            => st_rst,
    clk            => st_clk,

    -- Streaming Sink
    snk_in_err     => rx_adapt_sosi.err(c_tech_tse_error_w-1 DOWNTO 0),  -- preserve error field from TSE MAC stream
    snk_in         => rx_adapt_sosi,
    snk_out        => rx_adapt_siso,
    
    -- Streaming Source
    src_in         => rx_crc_siso,
    src_out        => rx_crc_sosi,        -- replaced CRC word by snk_in_err, so CRC word /=0 indicates any TSE error
    src_out_err    => OPEN                -- flag snk_in_err/=0 at src_out.eop
  );
  
  ------------------------------------------------------------------------------
  -- RX : Strip the option words from the IPv4 header
  ------------------------------------------------------------------------------
  gen_ihl20: IF g_ihl20 GENERATE
    u_ihl20 : ENTITY work.eth_ihl_to_20
    PORT MAP (
      rst            => st_rst,
      clk            => st_clk,

      -- Streaming Sink
      snk_in         => rx_crc_sosi,
      snk_out        => rx_crc_siso,
      
      -- Streaming Source
      src_in         => rx_ihl20_siso,
      src_out        => rx_ihl20_sosi
    );
  END GENERATE;
  
  no_ihl20: IF not g_ihl20 GENERATE
    rx_ihl20_sosi <= rx_crc_sosi;
    rx_crc_siso   <= rx_ihl20_siso;
  END GENERATE;
  
  
  ------------------------------------------------------------------------------
  -- RX : For IP verify IP header checksum
  ------------------------------------------------------------------------------
  
  u_rx_frame : ENTITY work.eth_hdr
  GENERIC MAP (
    g_header_store_and_forward     => TRUE,
    g_ip_header_checksum_calculate => TRUE
  )
  PORT MAP (
    -- Clocks and reset
    rst             => st_rst,
    clk             => st_clk,
    
    -- Streaming Sink
    snk_in          => rx_ihl20_sosi,
    snk_out         => rx_ihl20_siso,
    
    -- Streaming Source
    src_in          => rx_hdr_siso,
    src_out         => rx_hdr_sosi,
    
    -- Header info
    hdr_status      => rx_hdr_status
  );
  
  
  ------------------------------------------------------------------------------
  -- Demux the UDP off-load traffic and the keep the other ETH traffic
  ------------------------------------------------------------------------------

  -- Put UDP off-load traffic on channel > 0
  -- Put other ETH    traffic on channel = 0 for further internal processing
  u_udp_channel : ENTITY work.eth_udp_channel
  PORT MAP (
    -- Clocks and reset
    rst            => st_rst,
    clk            => st_clk,
    
    -- Streaming Sink
    snk_in         => rx_hdr_sosi,
    snk_out        => rx_hdr_siso,
    
    -- Streaming Source with channel field
    src_in         => rx_channel_siso,
    src_out        => rx_channel_sosi,
    
    -- Demux control
    reg_demux      => reg_demux,
    hdr_status     => rx_hdr_status
  );
  
  -- Demultiplex channel 0 for internal handling and the other channels > 0 for external UDP off-load.
  u_rx_demux : ENTITY dp_lib.dp_demux
  GENERIC MAP (
    g_nof_output    => c_demux_nof_ports,
    g_combined      => c_demux_combined
  )
  PORT MAP (
    rst         => st_rst,
    clk         => st_clk,
    -- ST sinks
    snk_out     => rx_channel_siso,
    snk_in      => rx_channel_sosi,
    -- ST source
    src_in_arr  => demux_siso_arr,
    src_out_arr => demux_sosi_arr
  );
    
  -- Fixed local ETH port
  eth_rx_sosi       <= demux_sosi_arr(0);
  demux_siso_arr(0) <= eth_rx_siso;

  -- UDP offload ports
  gen_udp_rx_demux : FOR i IN 1 TO c_eth_nof_udp_ports GENERATE
    udp_rx_src_out_arr(i-1) <= demux_sosi_arr(i);
    demux_siso_arr(i)       <= udp_rx_src_in_arr(i-1);
  END GENERATE;
 
  
  ------------------------------------------------------------------------------
  -- ETH RX frame buffer
  ------------------------------------------------------------------------------
  
  u_rx_buffer : ENTITY work.eth_buffer
  PORT MAP (
    -- Clocks and reset
    rst             => st_rst,
    clk             => st_clk,
    
    -- Streaming Sink
    snk_in          => eth_rx_sosi,
    snk_out         => eth_rx_siso,
    
    -- Streaming Source
    -- . The src_rd, src_ack and src_done act instead of src_in.ready to have src_in ready per frame instead of per data word
    src_rd          => rx_frame_rd,    -- request frame pulse
    src_ack         => rx_frame_ack,   -- acknowledge request
    src_done        => rx_frame_done,  -- signal frame received
    src_out         => rx_frame_sosi
  );  

    
  ------------------------------------------------------------------------------
  -- ETH RX frame monitor
  ------------------------------------------------------------------------------
  
  -- Extract total header
  u_rx_hdr_info : ENTITY work.eth_hdr
  GENERIC MAP (
    g_header_store_and_forward     => FALSE,
    g_ip_header_checksum_calculate => FALSE
  )
  PORT MAP (
    -- Clocks and reset
    rst             => st_rst,
    clk             => st_clk,
    
    -- Streaming Sink
    snk_in          => rx_frame_sosi,
    
    -- Header info
    hdr_words_arr   => rx_frame_hdr_words_arr,
    hdr_fields      => rx_frame_hdr_fields,
    hdr_status      => rx_frame_hdr_status
  );
  
  -- Extract CRC word (enumerate: 0=OK, >0 AND odd = Error)
  u_rx_crc_word : ENTITY work.eth_crc_word
  PORT MAP (
    rst            => st_rst,
    clk            => st_clk,
    
    -- Streaming Sink
    snk_in         => rx_frame_sosi,
    
    -- CRC word
    crc_word       => rx_frame_crc_word,
    crc_word_val   => OPEN
  );
  
  u_mm_reg_frame : ENTITY work.eth_mm_reg_frame
  PORT MAP (
    -- Clocks and reset
    rst             => st_rst,
    clk             => st_clk,
    
    -- Inputs need for the frame register
    hdr_fields      => rx_frame_hdr_fields,
    hdr_status      => rx_frame_hdr_status,
    erc_word        => rx_frame_crc_word,
    reg_config      => reg_config,
    
    -- Frame register
    reg_frame       => reg_frame
  );

    
  ------------------------------------------------------------------------------
  -- ETH Control
  ------------------------------------------------------------------------------
  
  u_control : ENTITY work.eth_control
  PORT MAP (
    -- Clocks and reset
    rst               => st_rst,
    clk               => st_clk,
    
    -- Control register
    reg_config        => reg_config,
    reg_control       => reg_control,
    reg_continue_wr   => reg_continue_wr,
    reg_status        => reg_status,
    reg_status_wr     => reg_status_wr,
        
    -- Streaming sink Rx frame
    rcv_rd            => rx_frame_rd,
    rcv_ack           => rx_frame_ack,
    rcv_done          => rx_frame_done,
    rcv_in            => rx_frame_sosi,
    rcv_hdr_words_arr => rx_frame_hdr_words_arr,
    rcv_hdr_status    => rx_frame_hdr_status,

    -- Streaming source Tx frame
    xmt_in            => eth_tx_siso,
    xmt_out           => eth_tx_sosi,
    
    -- MM frame memory    
    mem_in            => mem_in,
    mem_out           => mem_out
  );
  
    
  ------------------------------------------------------------------------------
  -- TX : Mux UDP
  ------------------------------------------------------------------------------
   
  -- Fixed local ETH
  eth_tx_siso     <= mux_siso_arr(0);
  mux_sosi_arr(0) <= eth_tx_sosi;

  -- UDP offload ports
  gen_udp_tx_mux : FOR i IN 1 TO c_eth_nof_udp_ports GENERATE
    udp_tx_snk_out_arr(i-1) <= mux_siso_arr(i);
    mux_sosi_arr(i)         <= udp_tx_snk_in_arr(i-1);
  END GENERATE;
  
  -- Multiplex the two input streams on to the single ETH stream
  u_tx_mux : ENTITY dp_lib.dp_mux
  GENERIC MAP (
    g_data_w          => c_eth_data_w,
    g_empty_w         => c_eth_empty_w,
    g_in_channel_w    => 1,
    g_error_w         => 1,
    g_use_empty       => TRUE,
    g_use_in_channel  => FALSE,
    g_use_error       => FALSE,
    g_nof_input       => c_mux_nof_ports,
    g_use_fifo        => FALSE,
    g_fifo_size       => array_init(1024, c_mux_nof_ports),  -- input FIFOs are not used, but generic must match g_nof_input
    g_fifo_fill       => array_init(   0, c_mux_nof_ports)   -- input FIFOs are not used, but generic must match g_nof_input
  )
  PORT MAP (
    rst         => st_rst,
    clk         => st_clk,
    -- ST sinks
    snk_out_arr => mux_siso_arr,     -- OUT = request to upstream ST source
    snk_in_arr  => mux_sosi_arr,
    -- ST source
    src_in      => tx_mux_siso,  -- IN  = request from downstream ST sink
    src_out     => tx_mux_sosi
  );  
  
  ------------------------------------------------------------------------------
  -- TX : For IP insert IP header checksum
  ------------------------------------------------------------------------------
  
  u_tx_frame : ENTITY work.eth_hdr
  GENERIC MAP (
    g_header_store_and_forward     => TRUE,
    g_ip_header_checksum_calculate => TRUE
  )
  PORT MAP (
    -- Clocks and reset
    rst             => st_rst,
    clk             => st_clk,
    
    -- Streaming Sink
    snk_in          => tx_mux_sosi,
    snk_out         => tx_mux_siso,
    
    -- Streaming Source
    src_in          => tx_hdr_siso,
    src_out         => tx_hdr_sosi
  );
  
  
  ------------------------------------------------------------------------------
  -- TSE MAC
  ------------------------------------------------------------------------------
  tx_hdr_siso <= tse_tx_siso;
  tse_tx_sosi <= func_dp_stream_error_set(tx_hdr_sosi, 0);   -- set err field (value 0 for OK)
  
  tse_tx_mac_in.crc_fwd <= '0';  -- when '0' then TSE MAC generates the TX CRC field

  u_tech_tse : ENTITY tech_tse_lib.tech_tse
  GENERIC MAP (
    g_technology   => g_technology,
    g_ETH_PHY      => g_ETH_PHY
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst         => mm_rst,
    mm_clk         => mm_clk,
    eth_clk        => eth_clk,
    tx_snk_clk     => st_clk,
    rx_src_clk     => st_clk,
    cal_rec_clk    => cal_rec_clk,
    -- Memory Mapped Slave
    mm_sla_in      => tse_sla_in,
    mm_sla_out     => tse_sla_out,
    -- MAC transmit interface
    -- . ST sink
    tx_snk_in      => tse_tx_sosi,
    tx_snk_out     => tse_tx_siso,
    -- . MAC specific
    tx_mac_in      => tse_tx_mac_in,
    tx_mac_out     => tse_tx_mac_out,  -- OPEN
    -- MAC receive interface
    -- . ST Source
    rx_src_in      => tse_rx_siso,
    rx_src_out     => tse_rx_sosi,
    -- . MAC specific
    rx_mac_out     => tse_rx_mac_out,
    -- PHY interface
    eth_txp        => eth_txp,
    eth_rxp        => eth_rxp,
    -- LED interface
    tse_led        => tse_led
  );
  
END str;
