--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;

ENTITY tr_xaui_align_dly IS    
  GENERIC(
    g_sim : BOOLEAN
  );              
  PORT (   
    tx_clk                    : IN STD_LOGIC;
    tx_rst                    : IN STD_LOGIC;

    a_rx_channelaligned       : IN STD_LOGIC;
    txc_rx_channelaligned_dly : OUT STD_LOGIC
  
  );
END tr_xaui_align_dly;


ARCHITECTURE rtl OF tr_xaui_align_dly IS  
  
  -- FSM to delay txc_rx_channelaligned (txc_rx_channelaligned_dly is connected to tx_siso.ready)
  CONSTANT c_align_dly_cnt             : NATURAL := sel_a_b(g_sim, 50, 150000000); -- about 1 second on hw
  CONSTANT c_align_dly_cnt_w           : NATURAL := ceil_log2(c_align_dly_cnt);

  TYPE t_state_enum IS (s_init, s_aligned, s_aligned_dly);

  SIGNAL state                         : t_state_enum;
  SIGNAL nxt_state                     : t_state_enum; 

  SIGNAL txc_rx_channelaligned         : STD_LOGIC;

  SIGNAL txc_align_dly_cycle_cnt       : STD_LOGIC_VECTOR(c_align_dly_cnt_w-1 DOWNTO 0);
  SIGNAL nxt_txc_align_dly_cycle_cnt   : STD_LOGIC_VECTOR(c_align_dly_cnt_w-1 DOWNTO 0);
 
  SIGNAL i_txc_rx_channelaligned_dly   : STD_LOGIC;
  SIGNAL nxt_txc_rx_channelaligned_dly : STD_LOGIC;

BEGIN 

  txc_rx_channelaligned_dly <= i_txc_rx_channelaligned_dly;

  u_async_txc_rx_channelaligned: ENTITY common_lib.common_async
  GENERIC MAP(
    g_rst_level => '0'
  )
  PORT MAP(
    rst  => tx_rst,
    clk  => tx_clk,
    din  => a_rx_channelaligned,
    dout => txc_rx_channelaligned
  );

  -- Assert txc_rx_channelaligned: basically indicates whether or not the RX of the receiving
  -- XAUI core (NOT this one) is assumed to be channelaligned.
  -- We assume the RX connected to our TX will be channel aligned within txc_align_dly_cycle_cnt cycles
  -- after 'our own' RX has asserted channelaligned. This is important because txc_tx_channelaligned_dly
  -- connects directly to tx_siso.ready.
  p_tx_clk : PROCESS(tx_rst, tx_clk)
  BEGIN
    IF tx_rst='1' THEN
      state                     <= s_init;  
      txc_align_dly_cycle_cnt   <= (OTHERS=>'0');
      i_txc_rx_channelaligned_dly <= '0';
    ELSIF rising_edge(tx_clk) THEN
      state                     <= nxt_state;
      txc_align_dly_cycle_cnt   <= nxt_txc_align_dly_cycle_cnt;
      i_txc_rx_channelaligned_dly <= nxt_txc_rx_channelaligned_dly;
    END IF;
  END PROCESS;

  p_state : PROCESS(state, txc_align_dly_cycle_cnt, txc_rx_channelaligned, i_txc_rx_channelaligned_dly)
  BEGIN  
    nxt_state                     <= state;  
    nxt_txc_align_dly_cycle_cnt   <= txc_align_dly_cycle_cnt;
    nxt_txc_rx_channelaligned_dly <= i_txc_rx_channelaligned_dly;
  CASE state IS
     
    WHEN s_init => --Wait until our own RX is channel aligned      
      IF txc_rx_channelaligned= '1' THEN
        nxt_state <= s_aligned;
      END IF;   
    
    WHEN s_aligned => -- We assume the connecting RX is aligned after this delay...
      nxt_txc_align_dly_cycle_cnt <= INCR_UVEC(txc_align_dly_cycle_cnt, 1); 
      IF txc_align_dly_cycle_cnt = TO_UVEC(c_align_dly_cnt, c_align_dly_cnt_w) THEN
        nxt_txc_rx_channelaligned_dly <= '1';
        nxt_state                     <= s_aligned_dly;
      END IF;

    WHEN s_aligned_dly => -- ... So now we can assume connected RX has aligned.
      -- Do nothing.
        
    END CASE;
  END PROCESS;


END rtl;

