import multiprocessing as mp
import time

###############################################################################
# Base classes
###############################################################################

class StreamPort(object):
    def __init__(self, name, mode, stream_index, data_width):
        self.name = name
        self.mode = mode
        self.stream_index = stream_index
        self.data_width = data_width
        self.pipe_end = None

    #'>' operator: Connect
    def __gt__(self, other):
        return self.__connect__(other)

    def __connect__(self, other):
        if not(self.pipe_end == None):
            # The pipe end is already assigned, this is the case for composite components that
            # have wired their entity port to internal instance ports. 
            # If self is a source port (mode='out'), this means that self.pipe_end is already occupied
            # with a pipe sink end from which this entity port receives from an internal source entity.
            # To make a connection between other and self, all we need to do is move the existing pipe's snk end
            # from self to other. The source end remains connected to self's internal instance.
            other.pipe_end = self.pipe_end
        elif not(other.pipe_end == None):
            # The pipe end is already assigned, this is the case for composite components that
            # have wired their entity port to internal instance ports. 
            # If other is a sink port (mode='in'), this means that other.pipe_end is already occupied
            # with a pipe source end from which this entity port send to an internal sink entity.
            # To make a connection between other and self, all we need to do is move the existing pipe's src end
            # from other to self. The sink end remains connected to other's internal instance.
            self.pipe_end = other.pipe_end
        else:
            snk_end, src_end = mp.Pipe(duplex=False)
            self.pipe_end  = src_end
            other.pipe_end = snk_end
        return Connection(self.name, other.name)

    def send(self, item):
        self.pipe_end.send(item) 

    def recv(self):
        while True:
            try:
                yield self.pipe_end.recv()
            except EOFError:
                break

class StreamPortArray(list):
    def __init__(self, name, mode, nof_streams, data_width):
        list.__init__(self)
        self.name = name
        self.mode = mode
        self.nof_streams = nof_streams
        self.data_width = data_width
        for stream_index in range(nof_streams):
            self.append(StreamPort(name, mode, stream_index, data_width))

    #'>' operator: Connect
    def __gt__(self, other):
        return self.__connect__(other)

    def __connect__(self, other):
        for stream_port_self, stream_port_other in zip(self,other):
            stream_port_self.__connect__(stream_port_other)
        return Connection(self.name, other.name)

    def recv(self):
        while True:
            try:
                items = []
                for stream_port in self:
                    items.append( stream_port.pipe_end.recv() )
                yield items
            except EOFError:
                break

class Connection():
    def __init__(self, src_component, snk_component):
        self.vhdl_assignments = ['niets']

    def generate(self, file_name):
        target_file = open(file_name, "w")
        # Write component instance
        target_file.write(self.vhdl_assignments[0])
        target_file.close()

class Component(mp.Process):
    def __init__(self, inst_nr, name, components=[], connections=[]):
        mp.Process.__init__(self)
        self.name = name
        self.inst_nr = inst_nr
        self.file_name = self.name
        self.ports = {}
        self.components = components
        self.connections = connections

    def start_components(self):
        if self.components == []:
            # This is not a composite component but a base component, start this instance itself
            self.start()
        else:
            # This is a composite component, start the internal instances
            for component in self.components:
                print 'Starting', component
                component.start_components()

    def terminate_components(self):
        if self.components == []:
            # This is not a composite component but a base component, terminate this instance itself
            self.terminate()
        else:
            # This is a composite component, terminate the internal instances
            for component in self.components:
                print 'Terminating', component
                component.terminate_components()

    def run_time(self, time_sec):
        self.start_components()
        time.sleep(time_sec)
        self.terminate_components()

    def generate(self):
        # write ASTRON header
        # []
        # write included libraries and packages
        # []
        # write entity declaration
        # []
        # write 'ARCHITECTURE str OF self.name IS'
        # []
        # write signal declarations
#        for component in self.components:
#            for port in component.signals:
#                signal.generate(self.file_name)

        # Write 'BEGIN'
        # []
        # write structural architecture (component instances)
        for component in self.components:
            component.generate(self.file_name)
        # write signal assignments
        for connection in self.connections:
            connection.generate(self.file_name)
        # write 'END str;'

    def __set_port__(self, name, mode, dimensions):
        """
        Mode:
        . 'in'
        . 'out'
        Dimension examples:
        . 32 = 1 stream, 32b (sosi)
        . (2,32) = 2 streams, 32b each (sosi_arr)
        . (2,4,32) = 2 buses, 4 streams/bus, 32b/stream (sosi_2arr)
        . (8,2,4,32) = 8 buses, 2 sub-buses/bus, 4 streams/sub-bus, 32b/stream (sosi_3arr)
        """
    def __set_port__(self, name, mode, dimensions):
        if isinstance(dimensions, int):
            # Single stream
            self.ports[name] = StreamPort(name, mode, stream_index=None, data_width=dimensions)
        elif len(dimensions)==2:
            # Multi stream
            self.ports[name] = StreamPortArray(name, mode, nof_streams=dimensions[0], data_width=dimensions[1])
        elif len(dimensions)==3:
            # Multi bus, multi stream
            self.ports[name] = StreamPortArray(name, mode, nof_buses=dimensions[0], nof_streams=dimensions[1], data_width=dimensions[2])

    def set_input(self, name, dimensions):
        self.__set_port__(name=name, mode='in', dimensions=dimensions)

    def set_output(self, name, dimensions):
        self.__set_port__(name=name, mode='out', dimensions=dimensions)

#    def generate(self, file_name):
#        target_file = open(file_name, "w")
#        # Write component instance
#        target_file.write(self.vhdl_instance)
#        target_file.close()

    #'>' operator: Connect
    def __gt__(self, other):
        return self.__connect__(other)

    def __connect__(self, other):
        for port_name_self, port_self in self.ports.iteritems():
            if port_self.mode == 'out':
                 for port_name_other,port_other in other.ports.iteritems():
                     if port_other.mode == 'in':
                         if type(port_self)==type(port_other):
                             port_self.__connect__(port_other)
        return Connection(self.name, other.name)


###############################################################################
# Component subclasses
###############################################################################

class SingleCounter(Component):
    def __init__(self, data_width=32, inst_nr=0):
        Component.__init__(self, inst_nr, name='SingleCounter')
        self.vhdl_instance = 'A string that contains the SingleCounter VHDL instance'
        self.set_output('src_out', data_width)
    def run(self):
        for value in xrange(10):
            item = value + self.inst_nr*100
            print '[Counter] src', self.inst_nr, ': sending', item
            # Offset sent value by 100 for each instance number increment
            self.ports['src_out'].send(item) 

class SinglePrinter(Component):
    def __init__(self, data_width=32, inst_nr=0):
        Component.__init__(self, inst_nr, name='SinglePrinter')
        self.vhdl_instance = 'A string that contains the SinglePrinter VHDL instance'
        self.set_input('snk_in', data_width)
    def run(self):
        for item in self.ports['snk_in'].recv():
            print '[Printer] snk', self.inst_nr, ': receiving', item    


class Counter(Component):
    def __init__(self, nof_streams, data_width=32, inst_nr=0):
        Component.__init__(self, inst_nr, name='Counter')
        self.vhdl_instance = 'A string that contains the Counter VHDL instance'
        self.set_output('src_out_arr', (nof_streams, data_width))
    def run(self):
        for value in xrange(10):
            for src in self.ports['src_out_arr']:
                item = value + src.stream_index*100
                print '[Counter] src', src.stream_index, ': sending', item
                # Offset sent value by 100 for each incremental stream
                src.send(item) 

class Through(Component):
    def __init__(self, nof_streams, data_width=32, inst_nr=0):
        Component.__init__(self, inst_nr, name='Through')
        self.vhdl_instance = 'A string that contains the Through VHDL instance'
        self.set_input('snk_in_arr', (nof_streams, data_width))
        self.set_output('src_out_arr', (nof_streams, data_width))
    def run(self):
        for items in self.ports['snk_in_arr'].recv():
            for item,src in zip(items, self.ports['src_out_arr']):
                print '[Through] snk', src.stream_index, ': forwarding', item, 'to src', src.stream_index
                src.send(item)

class Printer(Component):
    def __init__(self, nof_streams, data_width=32, inst_nr=0):
        Component.__init__(self, inst_nr, name='Printer')
        self.vhdl_instance = 'A string that contains the Printer VHDL instance'
        self.set_input('snk_in_arr', (nof_streams, data_width))
    def run(self):
        for items in self.ports['snk_in_arr'].recv():
            for item_index, item in enumerate(items):
                print '[Printer] snk', item_index, ': receiving', item


###############################################################################
# Main
###############################################################################

print
print '###############################################################################'
print '# Example 1: connect ports component-wise'
print '###############################################################################'
print

nof_streams = 3 

a = Counter(nof_streams)
b = Through(nof_streams)
c = Printer(nof_streams)

connections = [ a>b>c ]

d = Component(0, 'my_component_1', [a,b,c], connections)

d.run_time(1)
#d.generate()

print
print '###############################################################################'
print '# Example 2: connect ports array-wise'
print '###############################################################################'
print

nof_streams = 3 

a = Counter(nof_streams)
b = Through(nof_streams)
c = Printer(nof_streams)

connections = []
connections.append( a.ports['src_out_arr'] > b.ports['snk_in_arr' ] )
connections.append(                          b.ports['src_out_arr']   > c.ports['snk_in_arr'] )

d = Component(0, 'my_component_2', [a,b,c], connections)

d.run_time(1)
#d.generate()

print
print '###############################################################################'
print '# Example 3: connect ports stream-wise (with stream 0 / stream 1 swap)'
print '###############################################################################'
print

nof_streams = 3 

a = Counter(nof_streams)
b = Through(nof_streams)
c = Printer(nof_streams)

connections = []
connections.append( a.ports['src_out_arr'][0] > b.ports['snk_in_arr' ][0] )
connections.append( a.ports['src_out_arr'][1] > b.ports['snk_in_arr' ][1] )
connections.append( a.ports['src_out_arr'][2] > b.ports['snk_in_arr' ][2] )
connections.append(                             b.ports['src_out_arr'][1]   > c.ports['snk_in_arr'][0] ) # We're swapping streams 0 and 1 here
connections.append(                             b.ports['src_out_arr'][0]   > c.ports['snk_in_arr'][1] ) # We're swapping streams 0 and 1 here
connections.append(                             b.ports['src_out_arr'][2]   > c.ports['snk_in_arr'][2] )

d = Component(0, 'my_component_3', [a,b,c], connections)

d.run_time(1)
#d.generate()

print
print '###############################################################################'
print '# Example 4: Mix single and multi stream components'
print '###############################################################################'
print

nof_streams = 1

a = SingleCounter()
b = Through(nof_streams)
c = Printer(nof_streams)

connections = []
connections.append( a.ports['src_out'] > b.ports['snk_in_arr' ][0] )
connections.append(                      b.ports['src_out_arr']      > c.ports['snk_in_arr'] )

d = Component(0, 'my_component_4', [a,b,c], connections)

d.run_time(1)
#d.generate()

print
print '###############################################################################'
print '# Example 5: Creating multi-instance composites from single stream components'
print '# . We replace the Counter with a new component that instantiates 3 SingeCounters'
print '# . We replace the Printer with a new component that instantiates 3 SingePrinters'
print '###############################################################################'
print

nof_streams = 3
data_width = 32

# Create [nof_streams] SingleCounter instances
a_list = []
for i in range(nof_streams):
    a_list.append(SingleCounter(inst_nr=i))
# Create a new components wrapping these SingleCounter instances
a_multi = Component(0, 'a_multi', a_list)
# Declare the output of this new component
a_multi.set_output('src_out_arr', (nof_streams, data_width))
# Forward the SingleCounter outputs to the new component's output
for i in range(nof_streams):
    a_multi.components[i].ports['src_out'] > a_multi.ports['src_out_arr'][i]

# Unchanged: the Through component
b = Through(nof_streams)

# Create [nof_streams] SinglePrinter instances
c_list = []
for i in range(nof_streams):
    c_list.append(SinglePrinter(inst_nr=i))
# Create a new components wrapping these SingleCounter instances
c_multi = Component(0, 'c_multi', c_list)
# Declare the output of this new component
c_multi.set_input('snk_in_arr', (nof_streams, data_width))
# Forward the SingleCounter outputs to the new component's output
for i in range(nof_streams):
    c_multi.ports['snk_in_arr'][i] > c_multi.components[i].ports['snk_in']

# Continue as before (in the other examples)
connections = [ a_multi>b>c_multi ]

d = Component(0, 'my_component_5', [a_multi,b,c_multi], connections)

d.run_time(1)
#d.generate()


